unit FTPUtils;

interface

uses
  Classes,IniFiles,Contnrs,SyncObjs,

  LocalFileSystem,

  IdFTPList,IdFTPServer,IdIntercept,IdTCPConnection;

procedure GetDriveList(List:TStrings);overload;
function GetDriveList:TStrings;overload;

procedure GetFolderList(Path:string;List:TStrings);overload;
function GetFolderList(Path:string):TStrings;overload;

procedure GetFileList(Path:string;List:TStrings);overload;
function GetFileList(Path:string):TStrings;overload;

type
  TFTPPermissions=(ftpNone,ftpRead,ftpWrite);

  TFTPUserData=class;
  TFTPUserGroup=class;

  TLogEntryType=(lConnect,
                 lDisconnect,
                 lPermissionDeny,
                 lDownload,
                 lUpload,
                 lMakeFolder,
                 lMove,
                 lDeleteFile,
                 lRemoveFolder,
                 lBanIP);

  TLogEntrySet=set of TLogEntryType;

  TIPBan=record
    IP:string;
    BanTime:TDateTime;
  end;
  PIPBan=^TIPBan;

  TLogEntry=record
    Time:TDateTime;
    IP:string;
    EntryType:TLogEntryType;
    UserName,Description:string;
  end;
  PLogEntry=^TLogEntry;

  TFTPUserManager=class
  private
    FRootPath:string;
    FGroups:TObjectList;
    FAllowAnonymous: Boolean;
    FLogEntries,FIPBans,FSuspectIP:TStringList;
    FEntrySet:TLogEntrySet;
    FSection:TCriticalSection;
    FMaxUnsuccessfullConnections: Integer;
    FUseAntiBruteForce: Boolean;
    FCaseSensitiveAccounts: Boolean;
    FReceivedData: Int64;
    FSentData: Int64;
    FPermanentDelete: Boolean;
    FUpTrafficLimit,FDownTrafficLimit:Int64;
    FLastDownTrafficTick,FLastUpTrafficTick:TDateTime;
    FMaxUploadSpeed: Int64;
    FMaxDownloadSpeed: Int64;

    procedure SetRootPath(Value: string);
    procedure SetAllowAnonymous(const Value: Boolean);
    function GetGroup(Index: Integer): TFTPUserGroup;
    function GetGroupCount: Integer;

    function StrToLogEntry(s:string):TLogEntry;
    function LogEntryToStr(e:TLogEntry):string;
    function GetLogEntry(Index: Integer): TLogEntry;
    function GetLogEntryCount: Integer;
    function GetIPBan(Index: Integer): TIPBan;
    function GetIPBanCount: Integer;
    procedure SetMaxUnsuccessfullConnections(const Value: Integer);
    procedure SetUseAntiBruteForce(const Value: Boolean);
    procedure SetCaseSensitiveAccounts(const Value: Boolean);
    procedure SetReceivedData(const Value: Int64);
    procedure SetSentData(const Value: Int64);
    procedure SetPermanentDelete(const Value: Boolean);

    procedure ApplyTrafficLimitation(Bytes:Int64;Down:Boolean);
    procedure SetMaxDownloadSpeed(const Value: Int64);
    procedure SetMaxUploadSpeed(const Value: Int64);
  public
    constructor Create;

    procedure CleanPaths;

    function CreateUserData(Sender:TIdFTPServerThread;const AUsername,APassword:string):TFTPUserData;

    property RootPath:string read FRootPath write SetRootPath;

    property GroupCount:Integer read GetGroupCount;
    property Group[Index:Integer]:TFTPUserGroup read GetGroup;
    function AddGroup(Name:string):TFTPUserGroup;
    procedure DeleteGroup(Index:Integer);

    property AllowAnonymous:Boolean read FAllowAnonymous write SetAllowAnonymous;

    procedure ReadFromIniFile(IniFile:TCustomIniFile);
    procedure WriteToIniFile(IniFile:TCustomIniFile);

    procedure SaveLog(FileName:string);
    procedure LoadLog(FileName:string);
    procedure AddLogEntry(IP:string;EntryType:TLogEntryType;UserName,Description:string);
    property LogEntryCount:Integer read GetLogEntryCount;
    property LogEntry[Index:Integer]:TLogEntry read GetLogEntry;
    procedure ClearLog;
    procedure ClearOldLog(Age:TDateTime);
    property LogEntrySet:TLogEntrySet read FEntrySet write FEntrySet;

    procedure BanIP(AIP:string);
    procedure UnbanIP(AIP:string);
    property IPBanCount:Integer read GetIPBanCount;
    property IPBan[Index:Integer]:TIPBan read GetIPBan;

    property MaxUnsuccessfullConnections:Integer read FMaxUnsuccessfullConnections write SetMaxUnsuccessfullConnections;
    property UseAntiBruteForce:Boolean read FUseAntiBruteForce write SetUseAntiBruteForce;
    property CaseSensitiveAccounts:Boolean read FCaseSensitiveAccounts write SetCaseSensitiveAccounts;

    property SentData:Int64 read FSentData write SetSentData;
    property ReceivedData:Int64 read FReceivedData write SetReceivedData;

    property PermanentDelete:Boolean read FPermanentDelete write SetPermanentDelete;

    property MaxDownloadSpeed:Int64 read FMaxDownloadSpeed write SetMaxDownloadSpeed;
    property MaxUploadSpeed:Int64 read FMaxUploadSpeed write SetMaxUploadSpeed;

    procedure Lock;
    procedure UnLock;

    destructor Destroy;override;
  end;

  TFTPUser=class
  private
    FPassword: string;
    FUsername: string;
    FRealname: string;
    FActive: Boolean;

    procedure SetPassword(const Value: string);
    procedure SetRealname(const Value: string);
    procedure SetUsername(const Value: string);
    procedure SetActive(const Value: Boolean);
    function GetActive: Boolean;
  public
    constructor Create(AUsername,APassword,ARealname:string);

    property Username:string read FUsername write SetUsername;
    property Password:string read FPassword write SetPassword;
    property Realname:string read FRealname write SetRealname;
    property Active:Boolean read GetActive write SetActive;
  end;

  TFTPUserGroup=class
  private
    FReadRestrictions,FWriteRestrictions:TStringList;
    FGroupName: string;
    FOwner:TFTPUserManager;
    FUsers:TObjectList;
    FIsAdmin,FIsAnonymous: Boolean;

    function GetFolderPermission(Path:string):TFTPPermissions;
    procedure SetFolderPermission(Path:string;p:TFTPPermissions);
    function GetUserCount: Integer;
    function GetUser(Index: Integer): TFTPUser;
  public
    constructor Create(AOwner:TFTPUserManager;AGroupName:string;IsAdmin:Boolean=False;IsAnonymous:Boolean=False);

    property FolderPermission[Path:string]:TFTPPermissions read GetFolderPermission write SetFolderPermission;
    function CanChangeFolderPermission(Path:string):Boolean;

    property GroupName:string read FGroupName;

    property IsAdmin:Boolean read FIsAdmin;
    property IsAnonymous:Boolean read FIsAnonymous;

    property UserCount:Integer read GetUserCount;
    property User[Index:Integer]:TFTPUser read GetUser;
    procedure AddUser(Username,Password,Realname:string);
    procedure DeleteUser(Index:Integer);

    procedure WriteToIniFile(Section:string;IniFile:TCustomIniFile);
    procedure ReadFromIniFile(Section:string;IniFile:TCustomIniFile);

    destructor Destroy;override;
  end;

  TFTPUserData=class
  private
    FFileSystem:TLocalFileSystem;
    FUser:TFTPUser;
    FGroup:TFTPUserGroup;
    FUsername: string;
    FPassword: string;
    FIP: string;

    function FTPFileExists(FTPPath:string):Boolean;
    function FTPDirectoryExists(FTPPath:string):Boolean;
    function FTPParentDirectoryExists(FTPPath:string):Boolean;

    function FTPDirectoryHasPermission(FTPPath:string;Perm:TFTPPermissions):Boolean;
    function FTPParentDirectoryHasPermission(FTPPath:string;Perm:TFTPPermissions):Boolean;
    function ExtractParentPath(Path:string):string;

    function MergePath(s:string):string;

    procedure FTPCheck(Value:Boolean;ErrorMSG,TAG:string);

    function FTPToLocalDirectory(Path:string):string;
    function FTPToLocalFile(Path:string):string;
    function FTPToLocalParent(Path:string):string;
  public
    constructor Create(Sender:TIdFTPServerThread;AGroup:TFTPUserGroup;AUser:TFTPUser;AUserName,APassword:string);

    procedure DeleteFileEx(LocalPath:string);

    procedure GetDirectoryListing(const APath:string;ADirectoryListing:TIdFTPListItems);
    procedure ChangeDirectory(var VDirectory:string);
    procedure MakeDirectory(var VDirectory:string);
    procedure RemoveDirectory(var VDirectory:string);

    procedure GetFileSize(const AFilename:string;var VFileSize:Int64);
    procedure RetrieveFile(const AFileName:string;var VStream:TStream);
    procedure StoreFile(const AFileName:string;AAppend:Boolean;var VStream:TStream);
    procedure RenameFile(const ARenameFromFile,ARenameToFile:string);

    property Username:string read FUsername;
    property Password:string read FPassword;
    property User:TFTPUser read FUser;
    property Group:TFTPUserGroup read FGroup;
    property IP:string read FIP;

    procedure DeleteFile(const APathName:string);

    destructor Destroy;override;
  end;

  TServerTrafficIntercept=class(TIDServerIntercept)
  private
    FUserManager: TFTPUserManager;
    procedure SetUserManager(const Value: TFTPUserManager);
  public
    procedure Init;override;
    function Accept(AConnection:TComponent):TIdConnectionIntercept;override;
  published
    property UserManager:TFTPUserManager read FUserManager write SetUserManager;
  end;

  TTrafficConnectionIntercept=class(TIDConnectionIntercept)
  private
    FUserManager:TFTPUserManager;
  public
    constructor Create(AOwner:TComponent);override;

    procedure Receive(ABuffer:TStream);override;
    procedure Send(ABuffer:TStream);override;

    destructor Destroy;override;
  end;

const
  GDefaultEntrySet:TLogEntrySet=[lConnect,lDisconnect,lUpload,lMakeFolder,lMove,lDeleteFile,lRemoveFolder,lBanIP];

implementation

uses
  Windows,ShellAPI,SysUtils,StrUtils,Math,Forms;

procedure GetDriveList(List:TStrings);
var
  c:Char;
begin
  for c:='A' to 'Z' do
    if GetDriveType(PChar(c+':\'))<>1 then
      List.Add(c+':');
end;

function GetDriveList:TStrings;
begin
  Result:=TStringList.Create;
  GetDriveList(Result);
end;

procedure GetFolderList(Path:string;List:TStrings);overload;
var
  r:TSearchRec;
begin
  if FindFirst(Path+'*',faDirectory,r)=0 then begin
    repeat
      if (r.Attr and faDirectory=faDirectory) and (r.Name<>'.') and (r.Name<>'..') then
        List.Add(r.Name);
    until FindNext(r)<>0;
    FindClose(r);
  end;
end;

function GetFolderList(Path:string):TStrings;
begin
  Result:=TStringList.Create;
  GetFolderList(Path,Result);
end;

procedure GetFileList(Path:string;List:TStrings);overload;
var
  r:TSearchRec;
begin
  if FindFirst(Path+'*',faDirectory,r)=0 then begin
    repeat
      if r.Attr and faDirectory=0 then
        List.Add(r.Name);
    until FindNext(r)<>0;
    FindClose(r);
  end;
end;

function GetFileList(Path:string):TStrings;
begin
  Result:=TStringList.Create;
  GetFileList(Path,Result);
end;

{ TFTPUser }

constructor TFTPUser.Create;
begin
  inherited Create;
  FActive:=True;
  FUserName:=AUserName;
  FPassword:=APassword;
  FRealname:=ARealname;
  if ARealname='' then
    FRealname:=AUsername;
end;

function TFTPUser.GetActive: Boolean;
begin
  Result:=FActive and (FPassword<>'');
end;

procedure TFTPUser.SetActive(const Value: Boolean);
begin
  FActive := Value;
end;

procedure TFTPUser.SetPassword(const Value: string);
begin
  FPassword := Value;
end;

procedure TFTPUser.SetRealname(const Value: string);
begin
  FRealname := Value;
end;

procedure TFTPUser.SetUsername(const Value: string);
begin
  FUsername := Value;
  if FRealname='' then
    FRealName:=FUserName;
end;

{ TFTPUserManager }

function TFTPUserManager.AddGroup(Name: string): TFTPUserGroup;
var
  a:Integer;
begin
  Result:=nil;
  if Name='[Anonymous]' then begin
    AllowAnonymous:=True;
    Result:=Group[1];
    Exit;
  end;
  for a:=0 to FGroups.Count-1 do
    if Group[a].GroupName=Name then begin
      Result:=Group[a];
      Exit;
    end;
  if Name<>'' then
    Result:=Group[FGroups.Add(TFTPUserGroup.Create(Self,Name))];
end;

procedure TFTPUserManager.AddLogEntry(IP: string; EntryType: TLogEntryType;
  UserName, Description: string);
var
  p:PLogEntry;
begin
  FSection.Enter;
  try
    if EntryType in FEntrySet then begin
      New(p);
      p.Time:=Now;
      p.IP:=IP;
      p.EntryType:=EntryType;
      p.UserName:=UserName;
      p.Description:=Description;
      FLogEntries.AddObject(LogEntryToStr(p^),TObject(p));
//      ShowMessage(LogEntryToStr(p^));
    end;
  finally
    FSection.Leave;
  end;
end;

procedure TFTPUserManager.ApplyTrafficLimitation(Bytes: Int64;
  Down: Boolean);
const
  GUpdateIntervall=400;

  procedure Process(var LastTick:TDateTime;var TrafficCount,Counter:Int64;Max:Int64);
  var
    t:Double;
  begin
    if Max=0 then begin
      Counter:=Counter+Bytes;
      Exit;
    end;
    repeat
      t:=Now;
      Lock;
      try
        if t-LastTick>GUpdateIntervall/(3600*24*1000) then begin
          TrafficCount:=Max;
          LastTick:=t;
        end;
        if Bytes>TrafficCount then begin
          Bytes:=Bytes-TrafficCount;
          Counter:=Counter+TrafficCount;
          TrafficCount:=0;
        end else begin
          TrafficCount:=TrafficCount-Bytes;
          Counter:=Counter+Bytes;
          Bytes:=0;
        end;
      finally
        UnLock;
      end;
      if Bytes>0 then
        Sleep(GUpdateIntervall div 2);
    until Bytes=0;
  end;

begin
  if Down then
    Process(FLastDownTrafficTick,FDownTrafficLimit,FReceivedData,(FMaxDownloadSpeed div 1000)*GUpdateIntervall)
  else
    Process(FLastUpTrafficTick,FUpTrafficLimit,FSentData,(FMaxUploadSpeed div 1000)*GUpdateIntervall);
end;

procedure TFTPUserManager.BanIP(AIP: string);
var
  p:PIPBan;
  a:Integer;
begin
  FSection.Enter;
  try
    a:=FIPBans.IndexOf(AIP);
    if a=-1 then begin
      New(p);
      p.IP:=AIP;
      p.BanTime:=Now;
      FIPBans.AddObject(AIP,TObject(p));
    end else
      PIPBan(FIPBans.Objects[a]).BanTime:=Now;
    AddLogEntry('local',lBanIP,'Admin','IP '+AIP+' added to banlist');
  finally
    FSection.Leave;
  end;
end;

procedure TFTPUserManager.CleanPaths;
var
  a,b:Integer;
begin
  for a:=0 to FGroups.Count-1 do
    with Group[a] do begin
      for b:=FReadRestrictions.Count-1 downto 0 do begin
        if (Copy(FReadRestrictions[b],1,Length(FRootPath))<>FRootPath) or not DirectoryExists(FReadRestrictions[b]) then
          FReadRestrictions.Delete(b);
      end;
      for b:=FWriteRestrictions.Count-1 downto 0 do
        if (Copy(FWriteRestrictions[b],1,Length(FRootPath))<>FRootPath) or not DirectoryExists(FWriteRestrictions[b]) then
          FWriteRestrictions.Delete(b);
    end;
end;

procedure TFTPUserManager.ClearLog;
var
  a:Integer;
begin
  FSection.Enter;
  try
    for a:=0 to FLogEntries.Count-1 do
      Dispose(PLogEntry(FLogEntries.Objects[a]));
    FLogEntries.Clear;
  finally
    FSection.Leave;
  end;
end;

procedure TFTPUserManager.ClearOldLog(Age: TDateTime);
var
  a:Integer;
  p:PLogEntry;
begin
  FSection.Enter;
  try
    for a:=FLogEntries.Count-1 downto 0 do begin
      p:=PLogEntry(FLogEntries.Objects[a]);
      if (Now-p.Time>Age) or (Age=0) then begin
        Dispose(p);
        FLogEntries.Delete(a);
      end;
    end;
  finally
    FSection.Leave;
  end;
end;

constructor TFTPUserManager.Create;
var
  g:TFTPUserGroup;
begin
  inherited;
  FLastDownTrafficTick:=Now;
  FLastUpTrafficTick:=Now;
  FSection:=TCriticalSection.Create;
  FLogEntries:=TStringList.Create;
  FSuspectIP:=TStringList.Create;
  FIPBans:=TStringList.Create;
  RootPath:=AnsiReplaceStr(ExtractFilePath(Application.ExeName),'\','/');
  FGroups:=TObjectList.Create(True);
  g:=TFTPUserGroup.Create(Self,'[Administrator]',True);
  g.AddUser('Admin','','');
  FGroups.Add(g);
end;

function TFTPUserManager.CreateUserData(Sender: TIdFTPServerThread;
  const AUsername, APassword: string): TFTPUserData;
var
  a,b:Integer;
  s:string;
  t:Boolean;

  function CompareStr(s1,s2:string):Boolean;
  begin
    if FCaseSensitiveAccounts then
      Result:=s1=s2
    else
      Result:=AnsiUpperCase(s1)=AnsiUpperCase(s2);
  end;

begin
  s:=Sender.Connection.Socket.Binding.PeerIP;
  FSection.Enter;
  try
    t:=FIPBans.IndexOf(s)>-1;
    if t then
      AddLogEntry(s,lPermissionDeny,AUsername,'Connection refused, IP is in banlist');
  finally
    FSection.Leave;
  end;
  Result:=nil;
  if t then
    Exit;
  if FAllowAnonymous and CompareStr(AUsername,'Anonymous') then begin
    Result:=TFTPUserData.Create(Sender,Group[1],nil,AUsername,APassword);
    Exit;
  end;
  for a:=0 to FGroups.Count-1 do
    with TFTPUserGroup(FGroups[a]) do
      for b:=0 to FUsers.Count-1 do
        with TFTPUser(FUsers[b]) do
          if Active and CompareStr(AUsername,Username) and (APassword=Password) then begin
            Result:=TFTPUserData.Create(Sender,TFTPUserGroup(FGroups[a]),TFTPUser(FUsers[b]),AUsername,APassword);
            Exit;
          end;
  if not Assigned(Result) then
    AddLogEntry(s,lPermissionDeny,AUsername,'Invalid password : "'+APassword+'"');
  if UseAntiBruteForce and not Assigned(Result) then begin
    FSection.Enter;
    try
      a:=FSuspectIP.IndexOf(s);
      if a=-1 then
        FSuspectIP.AddObject(s,TObject(1))
      else begin
        FSuspectIP.Objects[a]:=TObject(Integer(FSuspectIP.Objects[a])+1);
        if Integer(FSuspectIP.Objects[a])>=FMaxUnsuccessfullConnections then
          BanIP(s);
      end;
    finally
      FSection.Leave;
    end;
  end;
end;

procedure TFTPUserManager.DeleteGroup(Index: Integer);
begin
  with Group[Index] do begin
    if IsAnonymous then begin
      AllowAnonymous:=False;
      Exit;
    end;
    if IsAdmin then
      Exit;
    FGroups.Delete(Index);
  end;
end;

destructor TFTPUserManager.Destroy;
var
  a:Integer;
begin
  FSection.Destroy;
  FSuspectIP.Destroy;
  for a:=0 to FIPBans.Count-1 do
    Dispose(PIPBan(FIPBans.Objects[a]));
  FIPBans.Destroy;
  for a:=0 to FLogEntries.Count-1 do
    Dispose(PLogEntry(FLogEntries.Objects[a]));
  FLogEntries.Destroy;
  FGroups.Destroy;
  inherited;
end;

function TFTPUserManager.GetGroup(Index: Integer): TFTPUserGroup;
begin
  Result:=TFTPUserGroup(FGroups[Index]);
end;

function TFTPUserManager.GetGroupCount: Integer;
begin
  Result:=FGroups.Count;
end;

function TFTPUserManager.GetIPBan(Index: Integer): TIPBan;
begin
  FSection.Enter;
  try
    Result:=PIPBan(FIPBans.Objects[Index])^;
  finally
    FSection.Leave;
  end;
end;

function TFTPUserManager.GetIPBanCount: Integer;
begin
  FSection.Enter;
  try
    Result:=FIPBans.Count;
  finally
    FSection.Leave;
  end;
end;

function TFTPUserManager.GetLogEntry(Index: Integer): TLogEntry;
begin
  FSection.Enter;
  try
    Result:=PLogEntry(FLogEntries.Objects[Index])^;
  finally
    FSection.Leave;
  end;
end;

function TFTPUserManager.GetLogEntryCount: Integer;
begin
  FSection.Enter;
  try
    Result:=FLogEntries.Count;
  finally
    FSection.Leave;
  end;
end;

procedure TFTPUserManager.LoadLog(FileName: string);
var
  a:Integer;
  p:PLogEntry;
begin
  FSection.Enter;
  try
    ClearLog;
    if FileExists(FileName) then begin
      FLogEntries.LoadFromFile(FileName);
      for a:=0 to FLogEntries.Count-1 do begin
        New(p);
        p^:=StrToLogEntry(FLogEntries[a]);
        FLogEntries.Objects[a]:=TObject(p);
      end;
    end;
  finally
    FSection.Leave;
  end;
end;

procedure TFTPUserManager.Lock;
begin
  FSection.Enter;
end;

function TFTPUserManager.LogEntryToStr(e: TLogEntry): string;
begin
  Result:=IntToStr(Integer(e.EntryType))+'['+FormatDateTime('dd/mm/ee - hh:nn:ss',e.Time)+'] user '+e.IP+' as "'+e.UserName+'" : '+e.Description;
end;

procedure TFTPUserManager.ReadFromIniFile(IniFile: TCustomIniFile);
var
  a,b:Integer;
  l:TLogEntryType;
  p:PIPBan;
begin
  while FGroups.Count>1 do
    FGroups.Delete(1);
  FAllowAnonymous:=False;
  AllowAnonymous:=IniFile.ReadBool('General','AllowAnonymous',False);
  FMaxUnsuccessfullConnections:=IniFile.ReadInteger('General','MaxUnsuccessfullConnections',100);
  FUseAntiBruteForce:=IniFile.ReadBool('General','UseAntiBruteForce',True);
  FCaseSensitiveAccounts:=IniFile.ReadBool('General','CaseSensitiveLogin',False);
  FPermanentDelete:=IniFile.ReadBool('General','PermanentDelete',False);
  FMaxDownloadSpeed:=IniFile.ReadInteger('General','MaxDownloadSpeed',0);
  FMaxUploadSpeed:=IniFile.ReadInteger('General','MaxUploadSpeed',0);
  FSection.Enter;
  try
    for a:=0 to FIPBans.Count-1 do
      Dispose(PIPBan(FIPBans.Objects[a]));
    FIPBans.Clear;
    b:=IniFile.ReadInteger('General','BanListCount',0);
    for a:=0 to b-1 do begin
      New(p);
      p.IP:=IniFile.ReadString('General','Ban'+IntToStr(a)+'.IP','');
      p.BanTime:=IniFile.ReadFloat('General','Ban'+IntToStr(a)+'.BanTime',Now);
      if p.IP<>'' then
        FIPBans.AddObject(p.IP,TObject(p));
    end;
  finally
    FSection.Leave;
  end;
  for l:=Low(TLogEntryType) to High(TLogEntryType) do
    if IniFile.ReadBool('General','LogStorage'+IntToStr(Integer(l)),l in GDefaultEntrySet) then
      Include(FEntrySet,l)
    else
      Exclude(FEntrySet,l);
  b:=IniFile.ReadInteger('General','GroupCount',0);
  for a:=1 to b-1 do
    AddGroup(IniFile.ReadString('General','Group'+IntToStr(a)+'.Name','Group '+IntToStr(a))).ReadFromIniFile('Group'+IntToStr(a),IniFile);
  Group[0].ReadFromIniFile('Group0',IniFile);
  RootPath:=IniFile.ReadString('General','RootPath',RootPath);
end;

procedure TFTPUserManager.SaveLog(FileName: string);
begin
  FSection.Enter;
  try
    FLogEntries.SaveToFile(FileName);
  finally
    FSection.Leave;
  end;
end;

procedure TFTPUserManager.SetAllowAnonymous(const Value: Boolean);
begin
  if Value xor FAllowAnonymous then begin
    FAllowAnonymous := Value;
    if Value then
      FGroups.Insert(1,TFTPUserGroup.Create(Self,'[Anonymous]',False,True))
    else
      FGroups.Delete(1);
  end;
end;

procedure TFTPUserManager.SetCaseSensitiveAccounts(const Value: Boolean);
begin
  FCaseSensitiveAccounts := Value;
end;

procedure TFTPUserManager.SetMaxDownloadSpeed(const Value: Int64);
begin
  FMaxDownloadSpeed := Value;
end;

procedure TFTPUserManager.SetMaxUnsuccessfullConnections(
  const Value: Integer);
begin
  FMaxUnsuccessfullConnections := Value;
end;

procedure TFTPUserManager.SetMaxUploadSpeed(const Value: Int64);
begin
  FMaxUploadSpeed := Value;
end;

procedure TFTPUserManager.SetPermanentDelete(const Value: Boolean);
begin
  FPermanentDelete := Value;
end;

procedure TFTPUserManager.SetReceivedData(const Value: Int64);
begin
  FReceivedData := Value;
end;

procedure TFTPUserManager.SetRootPath(Value: string);
begin
  if not DirectoryExists(Value) then begin
    Value:=ExtractFilePath(Application.ExeName);
    Value:=AnsiReplaceStr(Copy(Value,1,Length(Value)-1),'\','/');
  end;
    //raise Exception.Create('"'+Value+'" : directory does not exist');
  FRootPath := Value;
  while (FRootPath<>'') and (FRootPath[Length(FRootPath)]='/') do
    FRootPath:=Copy(FRootPath,1,Length(FRootPath)-1);
end;

procedure TFTPUserManager.SetSentData(const Value: Int64);
begin
  FSentData := Value;
end;

procedure TFTPUserManager.SetUseAntiBruteForce(const Value: Boolean);
begin
  FUseAntiBruteForce := Value;
end;

function TFTPUserManager.StrToLogEntry(s: string): TLogEntry;
var
  a,b:Integer;
begin
  Result.EntryType:=TLogEntryType(StrToInt(Copy(s,1,1)));
  Result.Time:=StrToDate(Copy(s,3,8))+StrToTime(Copy(s,14,8));
  a:=Pos(' as "',s);
  b:=Pos('" : ',s);
  Result.IP:=Copy(s,29,a-29);
  Result.UserName:=Copy(s,a+5,b-a-5);
  Result.Description:=Copy(s,b+4,Length(s)-b-3);
end;

procedure TFTPUserManager.UnbanIP(AIP: string);
var
  a:Integer;
begin
  FSection.Enter;
  try
    a:=FIPBans.IndexOf(AIP);
    if a>-1 then begin
      Dispose(PIPBan(FIPBans.Objects[a]));
      FIPBans.Delete(a);
      AddLogEntry('local',lBanIP,'Admin','IP '+AIP+' removed from banlist');
    end;
  finally
    FSection.Leave;
  end;
end;

procedure TFTPUserManager.UnLock;
begin
  FSection.Leave;
end;

procedure TFTPUserManager.WriteToIniFile(IniFile: TCustomIniFile);
var
  a:Integer;
  l:TLogEntryType;
begin
  IniFile.WriteString('General','RootPath',RootPath);
  IniFile.WriteInteger('General','GroupCount',GroupCount);
  IniFile.WriteBool('General','AllowAnonymous',AllowAnonymous);
  IniFile.WriteInteger('General','MaxUnsuccessfullConnections',FMaxUnsuccessfullConnections);
  IniFile.WriteBool('General','CaseSensitiveLogin',FCaseSensitiveAccounts);
  IniFile.WriteBool('General','PermanentDelete',FPermanentDelete);
  IniFile.WriteInteger('General','MaxDownloadSpeed',FMaxDownloadSpeed);
  IniFile.WriteInteger('General','MaxUploadSpeed',FMaxUploadSpeed);
  FSection.Enter;
  try
    IniFile.WriteInteger('General','BanListCount',FIPBans.Count);
    for a:=0 to FIPBans.Count-1 do
      with PIPBan(FIPBans.Objects[a])^ do begin
        IniFile.WriteString('General','Ban'+IntToStr(a)+'.IP',IP);
        IniFile.WriteFloat('General','Ban'+IntToStr(a)+'.BanTime',BanTime);
      end;
  finally
    FSection.Leave;
  end;
  IniFile.WriteBool('General','UseAntiBruteForce',FUseAntiBruteForce);
  for l:=Low(TLogEntryType) to High(TLogEntryType) do
    IniFile.WriteBool('General','LogStorage'+IntToStr(Integer(l)),l in FEntrySet);
  for a:=0 to GroupCount-1 do begin
    IniFile.WriteString('General','Group'+IntToStr(a)+'.Name',Group[a].GroupName);
    Group[a].WriteToIniFile('Group'+IntToStr(a),IniFile);
  end;
end;

{ TFTPUserGroup }

procedure TFTPUserGroup.AddUser(Username, Password, Realname: string);
begin
  FUsers.Add(TFTPUser.Create(Username,Password,Realname));
end;

function TFTPUserGroup.CanChangeFolderPermission(Path: string): Boolean;
begin
  Result:={(Path<>FOwner.FRootPath) and }not FIsAdmin;
end;

constructor TFTPUserGroup.Create(AOwner: TFTPUserManager; AGroupName: string;
  IsAdmin, IsAnonymous: Boolean);
begin
  inherited Create;
  FOwner:=AOwner;
  FIsAdmin:=IsAdmin;
  FIsAnonymous:=IsAnonymous;
  FGroupName:=AGroupName;
  FReadRestrictions:=TStringList.Create;
  FWriteRestrictions:=TStringList.Create;
  FUsers:=TObjectList.Create(True);
end;

procedure TFTPUserGroup.DeleteUser(Index: Integer);
begin
  FUsers.Delete(Index);
end;

destructor TFTPUserGroup.Destroy;
begin
  FReadRestrictions.Destroy;
  FWriteRestrictions.Destroy;
  FUsers.Destroy;
  inherited;
end;

function TFTPUserGroup.GetFolderPermission(Path: string): TFTPPermissions;
begin
  Result:=ftpNone;
  if (Path<>'') and (Path[Length(Path)]<>'/') then
    Path:=Path+'/';//Copy(Path,1,Length(Path)-1);
  if FReadRestrictions.IndexOf(Path)>-1 then
    Result:=ftpRead;
  if FWriteRestrictions.IndexOf(Path)>-1 then
    Result:=ftpWrite;
  if FIsAdmin then
    Result:=ftpWrite;
end;

function TFTPUserGroup.GetUser(Index: Integer): TFTPUser;
begin
  Result:=TFTPUser(FUsers[Index]);
end;

function TFTPUserGroup.GetUserCount: Integer;
begin
  Result:=FUsers.Count;
end;

procedure TFTPUserGroup.ReadFromIniFile(Section: string;
  IniFile: TCustomIniFile);
var
  a,b:Integer;
begin
  FUsers.Clear;
  FReadRestrictions.Clear;
  FWriteRestrictions.Clear;
  b:=IniFile.ReadInteger(Section,'UserCount',0);
  for a:=0 to b-1 do begin
    AddUser(IniFile.ReadString(Section,'User'+IntToStr(a)+'.Username',''),
            IniFile.ReadString(Section,'User'+IntToStr(a)+'.Password',''),
            IniFile.ReadString(Section,'User'+IntToStr(a)+'.Realname',''));
    User[a].Active:=IniFile.ReadBool(Section,'User'+IntToStr(a)+'.Active',False);
  end;
  b:=IniFile.ReadInteger(Section,'ReadPermissionsCount',0);
  for a:=0 to b-1 do
    FReadRestrictions.Add(IniFile.ReadString(Section,'ReadPermission'+IntToStr(a),'@@@'));
  b:=IniFile.ReadInteger(Section,'WritePermissionsCount',0);
  for a:=0 to b-1 do
    FWriteRestrictions.Add(IniFile.ReadString(Section,'WritePermission'+IntToStr(a),'@@@'));
  if FIsAdmin and (FUsers.Count=0) then
    AddUser('Admin','','');
end;

procedure TFTPUserGroup.SetFolderPermission(Path: string;
  p: TFTPPermissions);
var
  a:Integer;
begin
  Assert(CanChangeFolderPermission(Path),'Can''t change permissions for "'+Path+'"');
  a:=FReadRestrictions.IndexOf(Path);
  if a>-1 then
    FReadRestrictions.Delete(a);
  a:=FWriteRestrictions.IndexOf(Path);
  if a>-1 then
    FWriteRestrictions.Delete(a);
  case p of
    ftpRead:FReadRestrictions.Add(Path);
    ftpWrite:FWriteRestrictions.Add(Path);
  end;
end;

procedure TFTPUserGroup.WriteToIniFile(Section: string;
  IniFile: TCustomIniFile);
var
  a:Integer;
begin
  IniFile.WriteInteger(Section,'UserCount',UserCount);
  for a:=0 to UserCount-1 do begin
    IniFile.WriteString(Section,'User'+IntToStr(a)+'.Username',User[a].Username);
    IniFile.WriteString(Section,'User'+IntToStr(a)+'.Password',User[a].Password);
    IniFile.WriteString(Section,'User'+IntToStr(a)+'.Realname',User[a].Realname);
    IniFile.WriteBool(Section,'User'+IntToStr(a)+'.Active',User[a].Active);
  end;
  IniFile.WriteInteger(Section,'ReadPermissionsCount',FReadRestrictions.Count);
  for a:=0 to FReadRestrictions.Count-1 do
    IniFile.WriteString(Section,'ReadPermission'+IntToStr(a),FReadRestrictions[a]);
  IniFile.WriteInteger(Section,'WritePermissionsCount',FWriteRestrictions.Count);
  for a:=0 to FWriteRestrictions.Count-1 do
    IniFile.WriteString(Section,'WritePermission'+IntToStr(a),FWriteRestrictions[a]);
end;

{ TFTPUserData }

procedure TFTPUserData.ChangeDirectory(var VDirectory: string);
var
  s:string;
begin
  s:=MergePath(VDirectory);
  FTPCheck(FTPDirectoryHasPermission(s,ftpRead),'"'+s+'" : permission denied','CWD');
  FTPCheck(FTPDirectoryExists(s),'"'+s+'" : no such directory','CWD');
  FFileSystem.CurrentDirectory:='/'+FTPToLocalDirectory(s);
  VDirectory:=s;
end;

constructor TFTPUserData.Create(Sender: TIdFTPServerThread;
  AGroup: TFTPUserGroup; AUser: TFTPUser; AUserName,
  APassword: string);
var
  a:Integer;
begin
  inherited Create;
  a:=AGroup.FOwner.FSuspectIP.IndexOf(Sender.Connection.Socket.Binding.PeerIP);
  if a>-1 then
    AGroup.FOwner.FSuspectIP.Delete(a);
  FIP:=Sender.Connection.Socket.Binding.PeerIP;
  FGroup:=AGroup;
  FUser:=AUser;
  FUserName:=AUserName;
  FPassWord:=APassWord;
  FFileSystem:=TLocalFileSystem.Create(nil);
  FFileSystem.Connect;
  FFileSystem.CurrentDirectory:='/'+AGroup.FOwner.FRootPath;
  AGroup.FOwner.AddLogEntry(FIP,lConnect,AUserName,'Login with password "'+APassword+'"');
end;

procedure TFTPUserData.DeleteFile(const APathName: string);
begin
  FTPCheck(FTPParentDirectoryHasPermission(APathName,ftpWrite),'"'+APathName+'" : permission denied','Delete file');
  FTPCheck(FTPFileExists(APathName),'"'+APathName+'" : no such file','Delete file');
  FGroup.FOwner.AddLogEntry(FIP,lDeleteFile,Username,'Delete file: "'+FTPToLocalFile(APathName)+'"');
  DeleteFileEx(FTPToLocalFile(APathName));
end;

procedure TFTPUserData.DeleteFileEx(LocalPath: string);
var
  FOS:TSHFileOpStruct;
begin
  while (LocalPath<>'') and (LocalPath[Length(LocalPath)] in ['/','\']) do
    LocalPath:=Copy(LocalPath,1,Length(LocalPath)-1);
  if FGroup.FOwner.FPermanentDelete then begin
    if FileExists(LocalPath) then begin
      if not SysUtils.DeleteFile(LocalPath) then
        raise Exception.Create(SysErrorMessage(GetLastError));
    end else begin
      if not SysUtils.RemoveDir(LocalPath) then
        raise Exception.Create(SysErrorMessage(GetLastError));
    end;
  end else begin
    ZeroMemory(@FOS,SizeOf(FOS));
    with FOS do begin
      wFunc:=FO_DELETE;
      pFrom:=PChar(AnsiReplaceStr(LocalPath,'/','\')+#0#0);
      fFlags:=FOF_ALLOWUNDO or FOF_NOCONFIRMATION or FOF_SILENT;
    end;
    if ShFileOperation(FOS)<>0 then
      raise Exception.Create(SysErrorMessage(GetLastError));
  end;
end;

destructor TFTPUserData.Destroy;
begin
  FGroup.FOwner.AddLogEntry(FIP,lDisconnect,FUserName,'Disconnected');
  FFileSystem.Destroy;
  inherited;
end;

function TFTPUserData.ExtractParentPath(Path: string): string;
var
  a,b:Integer;
begin
  b:=0;
  for a:=Length(Path)-1 downto 1 do
    if Path[a]='/' then begin
      b:=a;
      Break;
    end;
  if b>0 then
    Result:=Copy(Path,1,b-1)
  else
    Assert(False,Path);
end;

procedure TFTPUserData.FTPCheck(Value: Boolean; ErrorMSG, TAG: string);
begin
  if not Value then begin
    FGroup.FOwner.AddLogEntry(FIP,lPermissionDeny,UserName,TAG+' : '+ErrorMSG);
    raise Exception.Create(TAG+' : '+ErrorMSG);
  end;
end;

function TFTPUserData.FTPDirectoryExists(FTPPath: string): Boolean;
begin
  Result:=FFileSystem.DirectoryExists('/'+FTPToLocalDirectory(FTPPath))
end;

function TFTPUserData.FTPDirectoryHasPermission(FTPPath: string;
  Perm: TFTPPermissions): Boolean;
begin
  Result:=FGroup.FolderPermission[FTPToLocalDirectory(FTPPath)]>=Perm;
end;

function TFTPUserData.FTPFileExists(FTPPath: string): Boolean;
begin
  Result:=FFileSystem.FileExists('/'+FTPToLocalFile(FTPPath));
end;

function TFTPUserData.FTPParentDirectoryExists(FTPPath: string): Boolean;
begin
  Result:=FFileSystem.DirectoryExists('/'+FTPToLocalParent(FTPPath)+'/')
end;

function TFTPUserData.FTPParentDirectoryHasPermission(FTPPath: string;
  Perm: TFTPPermissions): Boolean;
begin
  Result:=FGroup.FolderPermission[FTPToLocalParent(FTPPath)]>=Perm;
end;

function TFTPUserData.FTPToLocalDirectory(Path: string): string;
begin
  if (Path<>'') and (Path[1]='/') then begin
    while (Path<>'') and (Path[1]='/') do
      Path:=Copy(Path,2,Length(Path)-1);
    Result:=FGroup.FOwner.FRootPath+'/'+Path;
  end else
    Result:=Copy(FFileSystem.CurrentDirectory,2,Length(FFileSystem.CurrentDirectory)-1)+Path;
  if (Result<>'') and (Result[Length(Result)]<>'/') then
    Result:=Result+'/';
end;

function TFTPUserData.FTPToLocalFile(Path: string): string;
begin
  if (Path<>'') and (Path[1]='/') then begin
    while (Path<>'') and (Path[1]='/') do
      Path:=Copy(Path,2,Length(Path)-1);
    Result:=FGroup.FOwner.FRootPath+'/'+Path;
  end else
    Result:=Copy(FFileSystem.CurrentDirectory,2,Length(FFileSystem.CurrentDirectory)-1)+Path;
end;

function TFTPUserData.FTPToLocalParent(Path: string): string;
begin
  Path:=FTPToLocalDirectory(Path);
  Result:=ExtractParentPath(Path);
end;

procedure TFTPUserData.GetDirectoryListing(const APath: string;
  ADirectoryListing: TIdFTPListItems);
var
  a:Integer;
  s:string;
begin
  s:=MergePath(APath);
  FTPCheck(FTPDirectoryExists(s),'"'+s+'" : no such directory','LIST');
  FTPCheck(FTPDirectoryHasPermission(s,ftpRead),'"'+s+'" : permission denied','LIST');
  with FFileSystem.GetDirectoryContent('/'+FTPToLocalDirectory(APath)) do begin
    for a:=0 to Count-1 do
      if (not Items[a].Folder) or FTPDirectoryHasPermission(s+Items[a].FileName,ftpRead) then
        with ADirectoryListing.Add do begin
          FileName:=Items[a].FileName;
          Size:=Items[a].FileSize;
          ModifiedDate:=Items[a].DateModified;
          if Items[a].Folder then
            ItemType:=ditDirectory
          else
            ItemType:=ditFile;
        end;
    Destroy;
  end;
end;

procedure TFTPUserData.GetFileSize(const AFilename: string;
  var VFileSize: Int64);
var
  f:TFileStream;
begin
  FTPCheck(FTPParentDirectoryHasPermission(AFileName,ftpRead),'"'+AFileName+'" : permission denied','Get file size');
  FTPCheck(FTPFileExists(AFileName),'"'+AFileName+'" : no such file','Get file size');
  f:=TFileStream.Create(FTPToLocalFile(AFileName),fmOpenRead);
  VFileSize:=f.Size;
  f.Destroy;
end;

{function TFTPUserData.LocalToFtp(Path: string): string;
begin
  while (Path<>'') and (Path[1] in ['/','\']) do
    Path:=Copy(Path,2,Length(Path)-1);
  Assert(Copy(Path,1,Length(FGroup.FOwner.FRootPath))=FGroup.FOwner.FRootPath);
  Result:=Copy(Path,Length(FGroup.FOwner.FRootPath)-1,Length(Path));
end;}

procedure TFTPUserData.MakeDirectory(var VDirectory: string);
var
  a:Integer;
  s,t:string;
begin
  FTPCheck(FTPParentDirectoryHasPermission(VDirectory,ftpWrite),'"'+VDirectory+'" : permission denied','Make directory');
  FTPCheck(FTPParentDirectoryExists(VDirectory),'"'+VDirectory+'" : parent directory does not exist','Make directory');
  if not FTPDirectoryExists(VDirectory) then begin
    FGroup.FOwner.AddLogEntry(FIP,lMakeFolder,Username,'Make folder: "'+FTPToLocalDirectory(VDirectory)+'"');
    CreateDirectory(PChar(FTPToLocalDirectory(VDirectory)),nil);
    s:=FTPToLocalDirectory(VDirectory);
    t:=FTPToLocalParent(VDirectory);
    for a:=0 to FGroup.FOwner.GroupCount-1 do
      with FGroup.FOwner.Group[a] do
        if CanChangeFolderPermission(s) then
          FolderPermission[s]:=FolderPermission[t];
  end;
end;

function TFTPUserData.MergePath(s: string): string;
var
  a:Integer;
begin
  a:=Pos('\',s);
  if a=0 then
    Result:=s
  else begin
    Result:=Copy(s,a+1,Length(s)-a);
    s:=Copy(s,1,a-1);
    if Result='' then
      s:=s;
  end;
end;

procedure TFTPUserData.RemoveDirectory(var VDirectory: string);
var
  s:string;
begin
  s:=MergePath(VDirectory);
  FTPCheck(FTPParentDirectoryHasPermission(s,ftpWrite),'"'+s+'" : permission denied','Remove directory');
  FTPCheck(FTPDirectoryExists(s),'"'+s+'" : no such directory','Remove directory');
  FTPCheck(FTPDirectoryHasPermission(s,ftpWrite),'"'+s+'" : permission denied','Remove directory');
  with GetFileList(FTPToLocalDirectory(s)) do begin
    try
      FTPCheck(Count=0,'"'+s+'" : directory is not empty','Remove directory');
    finally
      Destroy;
    end;
  end;
  with GetFolderList(FTPToLocalDirectory(s)) do begin
    try
      FTPCheck(Count=0,'"'+s+'" : directory is not empty','Remove directory');
    finally
      Destroy;
    end;
  end;
  FGroup.FOwner.AddLogEntry(FIP,lRemoveFolder,Username,'Remove folder: "'+FTPToLocalDirectory(VDirectory)+'"');
  DeleteFileEx(FTPToLocalFile(s));//RemoveDirectory(PChar(FTPToLocal(s)));
  FGroup.FOwner.CleanPaths;
end;

procedure TFTPUserData.RenameFile(const ARenameFromFile,
  ARenameToFile: string);
var
  s,t:string;
  a,b:Integer;
begin
  FTPCheck(FTPParentDirectoryHasPermission(ARenameFromFile,ftpWrite),'"'+ARenameFromFile+'" : permission denied','Rename file');
  FTPCheck(FTPParentDirectoryHasPermission(ARenameToFile,ftpWrite),'"'+ARenameToFile+'" : permission denied','Rename file');
  if FTPDirectoryExists(ARenameFromFile) then begin
    FTPCheck(not FTPDirectoryExists(ARenameToFile),'"'+ARenameToFile+'" : directory already exists','Rename file');
    FGroup.FOwner.AddLogEntry(FIP,lMove,Username,'Move folder: "'+FTPToLocalDirectory(ARenameFromFile)+'" -> "'+FTPToLocalDirectory(ARenameToFile)+'"');
    MoveFile(PChar(FTPToLocalFile(ARenameFromFile)),PChar(FTPToLocalFile(ARenameToFile)));
    s:=FTPToLocalDirectory(ARenameFromFile);
    t:=FTPToLocalDirectory(ARenameToFile);
    for a:=0 to FGroup.FOwner.GroupCount-1 do begin
      with FGroup.FOwner.Group[a].FReadRestrictions do
        for b:=0 to Count-1 do
          if Copy(Strings[b],1,Length(s))=s then
            Strings[b]:=t+Copy(Strings[b],Length(s)+1,Length(Strings[b])-Length(s));
      with FGroup.FOwner.Group[a].FWriteRestrictions do
        for b:=0 to Count-1 do
          if Copy(Strings[b],1,Length(s))=s then
            Strings[b]:=t+Copy(Strings[b],Length(s)+1,Length(Strings[b])-Length(s));
    end;
  end else begin
    FTPCheck(FTPFileExists(ARenameFromFile),'"'+ARenameFromFile+'" : no such file or directory','Rename file');
    FTPCheck(not FTPFileExists(ARenameToFile),'"'+ARenameToFile+'" : file already exists','Rename file');
    FGroup.FOwner.AddLogEntry(FIP,lMove,Username,'Move file: "'+FTPToLocalFile(ARenameFromFile)+'" -> "'+FTPToLocalFile(ARenameToFile)+'"');
    MoveFile(PChar(FTPToLocalFile(ARenameFromFile)),PChar(FTPToLocalFile(ARenameToFile)));
  end;
end;

procedure TFTPUserData.RetrieveFile(const AFileName: string;
  var VStream: TStream);
begin
  FTPCheck(FTPParentDirectoryHasPermission(AFileName,ftpRead),'"'+AFileName+'" : permission denied','Retrieve file');
  FTPCheck(FTPFileExists(AFileName),'"'+AFileName+'" : no such file','Retrieve file');
  FGroup.FOwner.AddLogEntry(FIP,lDownload,Username,'Get file: "'+FTPToLocalFile(AFileName)+'"');
  VStream:=TFileStream.Create(FTPToLocalFile(AFileName),fmOpenRead);
  VStream.Position:=0;
end;

procedure TFTPUserData.StoreFile(const AFileName: string; AAppend: Boolean;
  var VStream: TStream);
begin
  FTPCheck(FTPParentDirectoryHasPermission(AFileName,ftpWrite),'"'+AFileName+'" : permission denied','Store file');
  FTPCheck(FTPParentDirectoryExists(AFileName),'"'+ExtractParentPath(AFileName)+'" : no such directory','Store file');
  FGroup.FOwner.AddLogEntry(FIP,lUpload,Username,'Store file: "'+FTPToLocalFile(AFileName)+'"');
  VStream:=TFileStream.Create(FTPToLocalFile(AFileName),fmOpenWrite or fmCreate);
  if AAppend then
    VStream.Position:=VStream.Size
  else
    VStream.Size:=0;
end;

{ TServerTrafficIntercept }

function TServerTrafficIntercept.Accept(
  AConnection: TComponent): TIdConnectionIntercept;
begin
  Result:=TTrafficConnectionIntercept.Create(AConnection);
  TTrafficConnectionIntercept(Result).FUserManager:=FUserManager;
end;

procedure TServerTrafficIntercept.Init;
begin

end;

procedure TServerTrafficIntercept.SetUserManager(
  const Value: TFTPUserManager);
begin
  FUserManager := Value;
end;

{ TTrafficConnectionIntercept }

constructor TTrafficConnectionIntercept.Create(AOwner: TComponent);
begin
  inherited;
  FConnection:=AOwner;
end;

destructor TTrafficConnectionIntercept.Destroy;
begin
  inherited;
end;

procedure TTrafficConnectionIntercept.Receive(ABuffer: TStream);
begin
  inherited;
  if Assigned(FUserManager) then begin
    if FUserManager.FMaxDownloadSpeed>0 then
      TIDTCPConnection(FConnection).RecvBufferSize:=Min(GRecvBufferSizeDefault,FUserManager.FMaxDownloadSpeed div 3)
    else
      TIDTCPConnection(FConnection).RecvBufferSize:=GRecvBufferSizeDefault;
    FUserManager.ApplyTrafficLimitation(ABuffer.Size,True);
  end;
end;

procedure TTrafficConnectionIntercept.Send(ABuffer: TStream);
begin
  inherited;
  if Assigned(FUserManager) then begin
    if FUserManager.FMaxUploadSpeed>0 then
      TIDTCPConnection(FConnection).SendBufferSize:=Min(GSendBufferSizeDefault,FUserManager.FMaxUploadSpeed div 3)
    else
      TIDTCPConnection(FConnection).SendBufferSize:=GSendBufferSizeDefault;
    FUserManager.ApplyTrafficLimitation(ABuffer.Size,False);
  end;
end;

end.
